"""Add state_reason to the model.

Revision ID: 229f6f273a56
Revises: 1a44272e8b4c
Create Date: 2016-10-20 10:18:03.019775

"""

# revision identifiers, used by Alembic.
revision = '229f6f273a56'
down_revision = '1a44272e8b4c'

from alembic import op
import sqlalchemy as sa


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.add_column('component_builds', sa.Column('state_reason', sa.String(), nullable=True))
    op.add_column('module_builds', sa.Column('state_reason', sa.String(), nullable=True))
    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('module_builds', 'state_reason')
    op.drop_column('component_builds', 'state_reason')
    ### end Alembic commands ###
